@@ -47,7 +47,8 @@ static GHashTable *window_hash = NULL;
     ((window)->priv->is_sticky           << 6) |        \
     ((window)->priv->is_hidden           << 7) |        \
     ((window)->priv->is_fullscreen       << 8) |        \
-    ((window)->priv->demands_attention   << 9) )
+    ((window)->priv->demands_attention   << 9) |        \
+    ((window)->priv->is_urgent           << 10) )
 
 struct _WnckWindowPrivate
 {
@@ -100,6 +101,7 @@ struct _WnckWindowPrivate
   guint is_hidden : 1;
   guint is_fullscreen : 1;
   guint demands_attention : 1;
+  guint is_urgent : 1;
 
   /* _NET_WM_STATE_HIDDEN doesn't map directly into an
    * externally-visible state (it determines the WM_STATE
@@ -125,6 +127,7 @@ struct _WnckWindowPrivate
   guint need_update_transient_for : 1;
   guint need_update_startup_id : 1;
   guint need_update_wmclass : 1;
+  guint need_update_wmhints : 1;
 };
 
 enum {
@@ -166,6 +169,9 @@ static void unqueue_update   (WnckWindow *window);
 static void queue_update     (WnckWindow *window);
 static void force_update_now (WnckWindow *window);
 
+static WnckWindow* find_last_transient_for (GList *windows,
+                                            Window xwindow);
+
 static gpointer parent_class;
 static guint signals[LAST_SIGNAL] = { 0 };
 
@@ -189,6 +195,7 @@ wnck_window_get_type (void)
         sizeof (WnckWindow),
         0,              /* n_preallocs */
         (GInstanceInitFunc) wnck_window_init,
+        NULL            /* value_table */
       };
 
       object_type = g_type_register_static (G_TYPE_OBJECT,
@@ -373,8 +380,10 @@ _wnck_window_create (Window      xwindow,
   _wnck_select_input (window->priv->xwindow,
                       WNCK_APP_WINDOW_EVENT_MASK);
 
-  window->priv->group_leader =
-    _wnck_get_group_leader (window->priv->xwindow);
+  /* Default the group leader to the window itself; it is set in
+   * update_wmhints() if a different group leader is specified.
+   */
+  window->priv->group_leader = window->priv->xwindow;
 
   window->priv->session_id =
     _wnck_get_session_id (window->priv->xwindow);
@@ -401,6 +410,7 @@ _wnck_window_create (Window      xwindow,
   window->priv->need_update_transient_for = TRUE;
   window->priv->need_update_startup_id = TRUE;
   window->priv->need_update_wmclass = TRUE;
+  window->priv->need_update_wmhints = TRUE;
   force_update_now (window);
 
   return window;
@@ -471,6 +481,14 @@ wnck_window_get_application  (WnckWindow *window)
   return window->priv->app;
 }
 
+WnckWindow*
+wnck_window_get_transient (WnckWindow *window)
+{
+  g_return_val_if_fail (WNCK_IS_WINDOW (window), None);
+
+  return wnck_window_get (window->priv->transient_for);
+}
+
 gulong
 wnck_window_get_group_leader (WnckWindow *window)
 {
@@ -589,6 +607,63 @@ wnck_window_get_window_type (WnckWindow *window)
   return window->priv->wintype;
 }
 
+/**
+ * wnck_window_set_window_type:
+ * @window: a #WnckWindow
+ * @wintype: the #WnckWindowType type hint
+ * 
+ * Sets the semantic type of the window.
+ * 
+ **/
+void
+wnck_window_set_window_type (WnckWindow *window, WnckWindowType wintype)
+{
+  Atom atom;
+
+  g_return_if_fail (WNCK_IS_WINDOW (window));
+
+  switch (wintype) {
+  case WNCK_WINDOW_NORMAL:
+    atom = _wnck_atom_get ("_NET_WM_WINDOW_TYPE_NORMAL");
+    break;
+  case WNCK_WINDOW_DESKTOP:
+    atom = _wnck_atom_get ("_NET_WM_WINDOW_TYPE_DESKTOP");
+    break;
+  case WNCK_WINDOW_DOCK:
+    atom = _wnck_atom_get ("_NET_WM_WINDOW_TYPE_DOCK");
+    break;
+  case WNCK_WINDOW_DIALOG:
+    atom = _wnck_atom_get ("_NET_WM_WINDOW_TYPE_DIALOG");
+    break;
+  case WNCK_WINDOW_MODAL_DIALOG:
+    atom = _wnck_atom_get ("_NET_WM_WINDOW_TYPE_MODAL_DIALOG");
+    break;
+  case WNCK_WINDOW_TOOLBAR:
+    atom = _wnck_atom_get ("_NET_WM_WINDOW_TYPE_TOOLBAR");
+    break;
+  case WNCK_WINDOW_MENU:
+    atom = _wnck_atom_get ("_NET_WM_WINDOW_TYPE_MENU");
+    break;
+  case WNCK_WINDOW_UTILITY:
+    atom = _wnck_atom_get ("_NET_WM_WINDOW_TYPE_UTILITY");
+    break;
+  case WNCK_WINDOW_SPLASHSCREEN:
+    atom = _wnck_atom_get ("_NET_WM_WINDOW_TYPE_SPLASHSCREEN");
+    break;
+  default:
+    return;
+  }
+  _wnck_error_trap_push ();
+
+  XChangeProperty (gdk_display,
+                   window->priv->xwindow, 
+                   _wnck_atom_get ("_NET_WM_WINDOW_TYPE"),
+		   XA_ATOM, 32, PropModeReplace,
+		   (guchar *)&atom, 1);
+
+  _wnck_error_trap_pop ();
+}
+
 /**
  * wnck_window_is_minimized:
  * @window: a #WnckWindow
@@ -607,23 +682,64 @@ wnck_window_is_minimized (WnckWindow *window)
 }
 
 /**
- * wnck_window_demands_attention:
+ * wnck_window_needs_attention:
  * @window: a #WnckWindow
  *
- * If the window is has the demands attention state set returns
- * %TRUE. This state may change anytime a state_changed signal gets
- * emitted.
+ * If the window needs attention returns %TRUE. This state may change
+ * anytime a state_changed signal gets emitted, as it can depend on flags
+ * such as the demands_attention and is_urgent hints.
  *
- * Return value: %TRUE if window is minimized
+ * Return value: %TRUE if window needs attention
  **/
 gboolean
-wnck_window_demands_attention (WnckWindow *window)
+wnck_window_needs_attention (WnckWindow *window)
 {
   g_return_val_if_fail (WNCK_IS_WINDOW (window), FALSE);
 
-  return window->priv->demands_attention;
+  return window->priv->demands_attention || window->priv->is_urgent;
 }
 
+/* Return whether one of the transients of @window needs attention */
+static gboolean
+transient_needs_attention (WnckWindow *window)
+{
+  GList *windows;
+  WnckWindow *transient;
+  
+  if (!WNCK_IS_WINDOW (window))
+    return FALSE;
+
+  windows = wnck_screen_get_windows_stacked (window->priv->screen);
+
+  transient = window;
+  while ((transient = find_last_transient_for (windows, transient->priv->xwindow)))
+    {
+      /* catch transient cycles */
+      if (transient == window)
+        return FALSE;
+
+      if (wnck_window_needs_attention (transient))
+        return TRUE;
+    }
+
+  return FALSE;
+}
+
+/**
+ * wnck_window_or_transient_needs_attention:
+ * @window: a #WnckWindow
+ *
+ * If the window or one of its transients needs attention returns %TRUE.
+ * This state may change anytime a state_changed signal gets emitted.
+ *
+ * Return value: %TRUE if window or one of its transients needs attention
+ **/
+gboolean
+wnck_window_or_transient_needs_attention (WnckWindow *window)
+{
+  return wnck_window_needs_attention (window) || 
+         transient_needs_attention (window);
+}
 
 gboolean
 wnck_window_is_maximized_horizontally (WnckWindow *window)
@@ -806,11 +922,12 @@ wnck_window_minimize                (WnckWindow *window)
 }
 
 void
-wnck_window_unminimize              (WnckWindow *window)
+wnck_window_unminimize              (WnckWindow *window,
+                                     guint32     timestamp)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  wnck_window_activate_transient (window);
+  wnck_window_activate_transient (window, timestamp);
 }
 
 void
@@ -1041,17 +1158,21 @@ wnck_window_unpin (WnckWindow *window)
 /**
  * wnck_window_activate:
  * @window: a #WnckWindow
+ * @timestamp: the X server timestamp of the user interaction event that
+ *             caused this call to occur
  *
  * Ask the window manager to make @window the active window.  The
  * window manager may choose to raise @window along with focusing it.
  **/
 void
-wnck_window_activate (WnckWindow *window)
+wnck_window_activate (WnckWindow *window,
+                      guint32     timestamp)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
   _wnck_activate (WNCK_SCREEN_XSCREEN (window->priv->screen),
-		  window->priv->xwindow);
+                  window->priv->xwindow,
+                  timestamp);
 }
 
 /**
@@ -1102,7 +1223,6 @@ wnck_window_is_most_recently_activated (WnckWindow *window)
   return (window == most_recently_activated_window);
 }
 
-
 static WnckWindow*
 find_last_transient_for (GList *windows,
                          Window xwindow)
@@ -1132,6 +1252,8 @@ find_last_transient_for (GList *windows,
 /**
  * wnck_window_activate_transient:
  * @window: a #WnckWindow
+ * @timestamp: the X server timestamp of the user interaction event that
+ *             caused this call to occur
  *
  * If @window has transients, activates the most likely transient
  * instead of the window itself. Otherwise activates @window.
@@ -1142,7 +1264,8 @@ find_last_transient_for (GList *windows,
  * 
  **/
 void
-wnck_window_activate_transient (WnckWindow *window)
+wnck_window_activate_transient (WnckWindow *window,
+                                guint32     timestamp)
 {
   GList *windows;
   WnckWindow *transient;
@@ -1170,23 +1293,19 @@ wnck_window_activate_transient (WnckWindow *window)
     }
 
   if (transient != NULL)
-    {
-      /* Raise but don't focus the main window */
-      XRaiseWindow (gdk_display, window->priv->xwindow);
-      /* then activate the transient */
-      wnck_window_activate (transient);
-    }
+    wnck_window_activate (transient, timestamp);
   else
-    {
-      wnck_window_activate (window);
-    }
+    wnck_window_activate (window, timestamp);
 }
 
 /**
- * wnck_window_transient_is_active:
+ * wnck_window_transient_is_most_recently_activated:
  * @window: a #WnckWindow
  *
- * Return whether @window or one of its transients has focus.  This
+ * Return whether one of the transients of @window is the most
+ * recently activated window.  See
+ * wnck_window_is_most_recently_activated() for a more complete
+ * description of what is meant by most recently activated.  This
  * function is needed because clicking on the tasklist once will
  * activate a transient instead of the window itself
  * (wnck_window_activate_transient), and clicking again should
@@ -1196,7 +1315,7 @@ wnck_window_activate_transient (WnckWindow *window)
  * 
  **/
 gboolean
-wnck_window_transient_is_active (WnckWindow *window)
+wnck_window_transient_is_most_recently_activated (WnckWindow *window)
 {
   GList *windows;
   WnckWindow *transient;
@@ -1207,13 +1326,13 @@ wnck_window_transient_is_active (WnckWindow *window)
   windows = wnck_screen_get_windows_stacked (window->priv->screen);
 
   transient = window;
-  while (transient = find_last_transient_for (windows, transient->priv->xwindow))
+  while ((transient = find_last_transient_for (windows, transient->priv->xwindow)))
     {
       /* catch transient cycles */
       if (transient == window)
         return FALSE;
 
-      if (wnck_window_is_active (transient))
+      if (wnck_window_is_most_recently_activated (transient))
         return TRUE;
     }
 
@@ -1574,14 +1693,18 @@ _wnck_window_process_property_notify (WnckWindow *window,
   else if (xevent->xproperty.atom ==
            _wnck_atom_get ("_NET_WM_ICON") ||
            xevent->xproperty.atom ==
-           _wnck_atom_get ("KWM_WIN_ICON") ||
-           xevent->xproperty.atom ==
-           _wnck_atom_get ("WM_HINTS"))
+           _wnck_atom_get ("KWM_WIN_ICON"))
     {
       _wnck_icon_cache_property_changed (window->priv->icon_cache,
                                          xevent->xproperty.atom);
       queue_update (window);
     }
+  else if (xevent->xproperty.atom ==
+  	   _wnck_atom_get ("WM_HINTS"))
+    {
+      window->priv->need_update_wmhints = TRUE;
+      queue_update (window);
+    }
 }
 
 void
@@ -1714,6 +1837,9 @@ update_state (WnckWindow *window)
       break;
     }
 
+  /* FIXME!!!!!!!!!! What in the world is this buggy duplicate of the code
+   * immediately above this for??!?!?
+   */
   switch (window->priv->wintype)
     {
     case WNCK_WINDOW_DESKTOP:
@@ -2046,6 +2172,39 @@ update_wmclass (WnckWindow *window)
                      &window->priv->res_name);
 }
 
+static void
+update_wmhints (WnckWindow *window)
+{
+  XWMHints *hints;
+
+  if (!window->priv->need_update_wmhints)
+    return;
+
+  _wnck_error_trap_push ();
+  hints = XGetWMHints (gdk_display, window->priv->xwindow);
+  _wnck_error_trap_pop ();
+
+  if (hints)
+    {
+      if ((hints->flags & IconPixmapHint) ||
+          (hints->flags & IconMaskHint))
+        _wnck_icon_cache_property_changed (window->priv->icon_cache,
+                                           _wnck_atom_get ("WM_HINTS"));
+
+      if (hints->flags & WindowGroupHint)
+          window->priv->group_leader = hints->window_group;
+
+      if (hints->flags & XUrgencyHint)
+        window->priv->is_urgent = TRUE;
+      else
+        window->priv->is_urgent = FALSE;
+
+      XFree (hints);
+    }
+
+  window->priv->need_update_wmhints = FALSE;
+}
+
 static void
 force_update_now (WnckWindow *window)
 {
@@ -2101,6 +2260,7 @@ force_update_now (WnckWindow *window)
 
   update_startup_id (window);    /* no side effects */
   update_wmclass (window);
+  update_wmhints (window);
   update_transient_for (window); /* wintype needs this to be first */
   update_wintype (window);
   update_wm_state (window);
